"""Database module

Main database module. Responsible for image metadata, grouping, similarity grouping.
Also responsible for providing GUI with transparent ways for generating thumbnails, loading image files and embedded metadata.

    begin                : Sat Dec 14 2002
    copyright            : (C) 2002 by Ricardo Niederberger Cabral
    email                : nieder|at|mail.ru
    Time-stamp:            <03/03/08 22:11:49 rnc>
"""
## ***************************************************************************
## *                                                                         *
## *   This program is free software; you can redistribute it and/or modify  *
## *   it under the terms of the GNU General Public License as published by  *
## *   the Free Software Foundation; either version 2 of the License, or     *
## *   (at your option) any later version.                                   *
## *                                                                         *
## *************************************************************************** 

try:
    import sys
    import os
    import traceback
    import md5                          # hashing filenames to generate thumbnails
    import random                       # for generating unique id's
    import imghdr                       # to determine image type
    import time                         # to test performance increase during development, and generating batch names
    import marshal                      # for loading and saving img databases
    import difflib                      # to match similar filenames
    from string import *
    # imgSeek imports
    import EXIF
    import Error
    import Settings
except:
    traceback.print_exc()
    print "Error importing the necessary python modules. Unable to continue."
    sys.exit()
try:
    from qt import *
except:
    pass
    Error.PrintTB("You system doesn't seem to have PyQT installed. Please install it before running this application. Please, see http://www.riverbankcomputing.co.uk/pyqt/download.php")

try:
    if int(QT_VERSION_STR[0]) >=3:   # guess qt version. TODO: remove it one day. Windows is now on separate source tree
        qtv=3
    else:
        qtv=2
except:
    qtv=3
    
Settings.qtv=qtv

try:
    import imgdb
except:
    traceback.print_exc()
    print "Warning: Unable to load the C++ extension \"imgdb.so\" module. Make sure you installed imgSeek correctly, and email any install bug to \"imgseek-devel@lists.sourceforge.net\"."
    sys.exit()
####### Optional modules
## IPTC from PIL
try:
    import IptcExtract
    Settings.hasIPTC=1
except:
    Settings.hasIPTC=0
    
if qtv==2:                              # PIL is needed to open unsopported image formats in qt2
    try:
        import Image
    except:
        print "Error importing PIL Image module"
        pass

####### Detect supported image formats
Settings.exif_ext=["jpeg","jpe","jpg"]
Settings.iptc_ext=["tif","tiff","jpeg","jpe","jpg"]
if qtv>=3:
    Settings.qt_ext=["jpeg","jpe","jpg","gif","bmp","png","xbm","pnm"]    
else:
    Settings.qt_ext=["gif","bmp","png","xbm"]
Settings.magick_ext=["tiff","tif","tiff","tga","targa","jpeg","jpe","jpg","gif","bmp","png","xbm","pnm",
"art","avs","cgm","cmyk","cut","dcm","dcx","dib","dpx","ept","fig","fpx","gplt","gray","hpgl","html","ico","jbig","jp2","jpc","mat","miff","mono",
"mng","otb","p7","pbm","pcd","pcx","pdf","pgm","pix","ppm","ps","rad","rgb","sct","sgi","sun","txt","xbm","xcf"]
Settings.pil_ext=["jpeg","jpe","jpg","gif","bmp","png","xbm","pnm"]

Settings.supported_ext=Settings.qt_ext

class ImgDB:
    def __init__(self):
        self.dbversion=7                # saved on the beginning of marshalled db files. For version db detection and handling
        self.fname=""                   # the filename of the currently opened database
        self.imgdb=imgdb
        # set supported extensions
        if imgdb.hasImageMagick():
            Settings.hasImMagick=1
            Settings.supported_ext=Settings.qt_ext+Settings.magick_ext
        else:
            Settings.hasImMagick=0
            Settings.magick_ext=[]
            Settings.supported_ext=Settings.qt_ext
        if Settings.hasPIL:
            Settings.supported_ext=Settings.qt_ext+Settings.pil_ext
        #### image metadata:
        self.imageParms=["Description","Dimensions","Filesize","Filename","Format","Collection","Author","Modify date","Database date"] # default meta keys all image should have
        self.blankMetaDict={}           # blank dictionary with available keys. To be copied later to every new db file
        for par in self.imageParms:
            self.blankMetaDict[par]=None
        self.dirty=0
        self.imgexts = [ 'jpeg', 'jpg', 'gif', 'png', 'rgb', 'pbm', 'pgm', 'ppm', 'tiff', 'tif', 'rast', 'xbm', 'bmp' ] # to help determining img format from extension
        #### Thumbnail stuff
        self.thdir=os.path.expanduser(os.path.join("~",".thumbnails","")) # path for storing thumbnails. Calculate it only once here.
        self.thdir2=os.path.expanduser(os.path.join("~",".thumbnails","normal","")) # path for storing thumbnails. Calculate it only once here.
        # just to make sure the thumbnail dir exists, so I dont need to be checking it over and over on self.getThumb()
        if not os.path.exists(self.thdir):
            os.mkdir(self.thdir)
        if not os.path.exists(self.thdir2):
            os.mkdir(self.thdir2)        
        #### Callbacks (lists of functions that should be called on changes            
        self.wnd=None
        self.app=None
        # this functions should accept a what map describing what changed on db
        self.cbs={}
        self.cbs["Batch"]=[]
        self.cbs["Dir"]=[]
        self.cbs["Group"]=[]
        self.cbs["SimGroup"]=[]
        self.cbs["Img"]=[]
        self.cbs["Volume"]=[]
        self.cbs["Meta"]=[]        
        self.cbs["Database"]=[]          # generic changes like, dbname, location. etc
        self.sysdirbmCb=[]        
        ### Init C++ db engine
        self.doNotSave=0                # flag to abort automatic db saving
        self.reset()
        imgdb.initDbase()
        ### Misc inits
        self.readonlyfields=["Filesize","Filename","Format","Dimensions"] # default readonly metadata
    def changedDB(self,what=None):
        """helper function, call it whenever something on db is changed and the GUI should know of

        @param what: describe the change
        @type what: dict
        """
        self.dirty=1
        if not what:
            print "Changes should have a reason."
            return
        try:
            for cb in self.cbs[what["scope"]]:
                cb(what) 
        except:
            traceback.print_exc()
            pass
    def remove_dead(self):
        """removes any non existant file on database """        
        #TODO: bug: calling it will remove files on removable volumes (imgs on cds won't be found and will be removed from db)
        #fix: volume attrib telling it's a removable vol ?
        remd=[]
        self.doNotSave=1        
        for fidx in self.img.keys():
            if not os.path.exists(self.img[fidx][0]):
                self.removeFile(fidx)
                remd.append(fidx)
        self.doNotSave=0                
        self.dirty=1
        self.changedDB({"scope":"Img","reason":"removed","subjects":remd})
        return len(remd)
    def rescan(self):
        """recalculate coefficients for all images on the database and regenerate the db buckets  """
        #TODO: store filesize to determine if a file should be rescanned or not ?
        #this feature doesn't seem to be needed, as its very unlikely that images will be changed, only removed.
        pass
    def reset(self):
        """ reset current db, should be called on File-> New """
        ## reseting all data structures
        self.img={}                     # key is id, value is [path,parent dir id, [group ids]]
        self.meta={}                    # key is id, value is dictionary of meta values
        self.dirs={}                    # key id, value is [0path,1-parent dir id,2-child dir ids,3-child files ids,4-child imgs filenames,5-description,6-volumeid]
        self.volumes={}                 # key id, value is [0-list of dir ids on this volume,1-base path,2-description,3-name]
        self.batches={}                 # key id, value is [name,contents]
        # contents types:
        # "SysDir", "Dir", "Group",  "SimGroup",  "Img", "Volume", "File"
        # a File id is the path
        self.groups={}# key is group id, value is [0-name,1-description,2-parent id, 3-child imgs id, 4-child groups 5-id]
        self.opened=0                   # db has been succesfully opened
        self.dirty=0                    # has db been changed since last change ?        
        self.openGroups=[]              # group ids with listvierw item expanded (so ui will set the listview state when it changed)
        self.openDirs=[]                # dir ids with listvierw item expanded (so ui will set the listview state when it changed)        
        self.simgroups={}# key is id value is [0-group name,1-main img id,2-list of similar images ids]
        self.metafields=self.imageParms[:] # history of entered metafields        
        self.textqueryhistory=[]        # key is text str, value is list of [field,value] pairs
        self.sysdirbm=[]                # browse by sysdir bookmark list. List of paths
        self.filenamedict={}            # key is filename value is id
        self.fullfilenamedict={}        # key is filename value is id        
        ###### Standard items on datastructures:
        # standard group where all images are added to if nothing else is specified
        self.groups[1]=["Orphan","Images which haven't been assigned to any other group.",-1,[],[]]
        self.volumes[1]=[[],os.sep,"Default volume","Local filesystem"]        
        self.batches[1]=["Temporary work batch",[]]
        self.kbsize=0
        self.curBatch=1                 # current batch being added.
        self.dirty=1
        self.imgdb.resetdb()
        self.changedDB({"scope":"Database","reason":"reset"})
    def opendb(self,fname):
        """open metadata db, losing current db. Will also tell imgdb C++ module to load corresponding image data

        @param fname: full path filename
        @type fname: string
        """
        self.fname=os.path.expanduser(fname)
        if not os.path.exists(self.fname):
            print "Starting with a new database."
            self.reset()
            return 0        
        print "Opening database file "+self.fname+" ..."
        if not imgdb.loaddb(self.fname+".img"):
            print "Error loading image database"
        rdbversion=1                    # database version read from db file
        try:                            # get db file size for statistical purposes
            self.kbsize=os.stat(self.fname).st_size
        except:
            self.kbsize=0
        try:
            f=open(self.fname,"rb")
            rdbversion=marshal.load(f)
            if rdbversion < 3:
                self.v1files=marshal.load(f)
                self.v1buckets=marshal.load(f)
                self.v1dirs=marshal.load(f)
                self.v1dirtree=marshal.load(f)
                self.v1groups=marshal.load(f)            
                if rdbversion>1:
                    self.v1meta=marshal.load(f)
            else: # load db > v3 < v5
                self.img=marshal.load(f)
                self.meta=marshal.load(f)
                if rdbversion < 5:
                    self.buckets=marshal.load(f)
                self.dirs=marshal.load(f)
                self.volumes=marshal.load(f)
                if rdbversion==3: zombies=marshal.load(f) # deprecated. Was removed on db version 4 
                self.groups=marshal.load(f)
                self.simgroups=marshal.load(f)
                self.metafields=marshal.load(f)
                self.textqueryhistory=marshal.load(f)
                self.openGroups=marshal.load(f)
            if rdbversion==4:
                self.sig=marshal.load(f)
            if rdbversion>=4:                
                self.batches=marshal.load(f)
                self.sysdirbm=marshal.load(f)
            if rdbversion==5:
                # fix list of groups each image belongs to
                print "Importing version 5 database"
            if rdbversion>=7:
                self.filenamedict=marshal.load(f)
                self.fullfilenamedict=marshal.load(f)
                self.openDirs=marshal.load(f)
            self.syncFilenames()
            self.syncFullFilenames()
            f.close()
            self.batches[1]=["Temporary work batch",[]]
            self.opened=1            
        except:
            traceback.print_exc()
            print "Error opening database, starting with an empty one."
            self.reset()
            return 0
        if rdbversion==1 or rdbversion==2:
            # import a dbversion 1 db and increment it to the next version so the next for catches it and do the proper imports for the next version            
            print "Importing version 2 image database to version 3. This will require imgSeek to rescan all images you had on your database."
            print "Please wait..."
            for dir in self.v1dirtree.keys():
                if not self.v1dirtree[dir]: # is root dir
                    self.addDir(1,dir,1,[],1)
            # import descriptions
            if rdbversion==2:
                for k in self.v1meta.keys():
                    try:
                        if self.v1meta[k]["Description"]:
                            for fid in self.meta.keys():
                                if find(self.v1meta[k]["Filename"],self.meta[fid]["Filename"]) != -1:
                                    self.meta[fid]["Description"]=self.v1meta[k]["Description"]
                    except:
                        pass
            self.dirty=1
        if rdbversion ==3 or rdbversion ==4: 
            print "Importing database from previous version. This will require imgSeek to rescan all images you had on your database."
            print "Adding %d images. Please wait..." % len(self.img.keys())
            for id in self.img.keys():
                if not imgdb.addImage(id,self.img[id][0],self.getThumbName(self.img[id][0]),not os.path.exists(self.getThumbName(self.img[id][0]))):
                    print "Error adding image:",self.img[id][0]

            self.dirty=1
        if rdbversion <7:
            self.syncFilenames()            
        print "Done."
        self.changedDB({"scope":"Database","reason":"opened","subject":fname})
        #TODO add dirbm to cb system
        for cb in self.sysdirbmCb: cb()        
        return 1
    def fixImgGroups(self,id):
        """ call to make sure the list of groups this img id belongs to is right

        @param id: image id to be fixed
        """
        #TODO: improve it. check if parents know all children and viceversa
        self.img[id][2]=[]
        for gid in self.groups.keys():
            if id in self.groups[gid][3]:
                self.img[id][2].append(gid)
    def changefname(self,nname):
        """ call to change db filename name"""
        self.fname=nname
        self.dirty=1
        self.changedDB({"scope":"Database","reason":"changedname","subject":nname})        
    def generateImageID(self):
        """ need to return something unique"""
        #TODO: is this random hack safe ? should I hash a resource somehow ? should I check if this ID already exists somewhere ? How much will extra checks slow things down ?
        try:
            a=random.randrange(999999999)
        except:
            try:
                a=random.randrange(99999999)
            except:                
                try:
                    a=random.randrange(9999999)
                except:
                    a=random.randrange(999999)                
        # 1 is a reserved id for default items
        while a==1: a=self.generateImageID()
        return a
    def savedb(self):
        """save all metadata/grouping/etc. Will also tell imgdb C++ module to save img coeffs.
        """
        if not imgdb.savedb(self.fname+".img"):
            print "Error saving image database"
        try:
            f=open(self.fname,"wb")
            marshal.dump(self.dbversion,f)
            marshal.dump(self.img,f)
            marshal.dump(self.meta,f)
            marshal.dump(self.dirs,f)
            marshal.dump(self.volumes,f)
            marshal.dump(self.groups,f)
            marshal.dump(self.simgroups,f)            
            marshal.dump(self.metafields,f)            
            marshal.dump(self.textqueryhistory,f)            
            marshal.dump(self.openGroups,f)
            marshal.dump(self.batches,f)            
            marshal.dump(self.sysdirbm,f)
            marshal.dump(self.filenamedict,f)
            marshal.dump(self.fullfilenamedict,f)
            marshal.dump(self.openDirs,f)            
            f.close()
        except:
            traceback.print_exc()            
            print "Error saving metadata."
            return 0
        self.dirty=0
        self.changedDB({"scope":"Database","reason":"saved","subject":self.fname})
        print "Database saved."        
        return 1
    def getImageFormat(self,fname):
        """determine image format

        @param fname: full filename
        @type fname: string
        @return: image format in lowercase
        @rtype: string
        """
        (base,ext) = os.path.splitext(fname)
        if ext[1:].lower() in self.imgexts:
            return ext[1:].lower()
        else:                           # fallback to imghdr
            try:
                return imghdr.what(fname)
            except:
                return "Unknown"
    def extIsImg(self,file,probe=0):
        """ tests if this file has a supported image format extension or is an image.

        @param file: full filename
        @type file: string
        @param probe: if true, makes it check even if file has no extension
        @type probe: int
        """
        ext=rfind(file,".")        
        if ext==-1:
            if not probe:
                return 0
            else:
                try:
                    return imghdr.what(file)
                except:
                    return None            
        ext=lower(file[ext+1:])        
        if ext in Settings.supported_ext:  # if it is a supported image filetype
            return ext
        if probe:
            try:
                return imghdr.what(file)
            except:
                return 0
        return 0
    def updateGroups(self,app,simthresd=50):
        """ groups by similarity the imgs id listed on targetids. 

        @param simthresd:  is the similarity threshold for choosing groups
        @type simthresd: integer
        """        
        solved=[]
        addDirProgressBar=QProgressDialog( "Creating groups, please wait...", "Abort", 10,app.mainWidget(), "addProgress", 1 )
        targetids=self.img.keys()        
        addDirProgressBar.setTotalSteps(len(targetids))
        addDirProgressBar.setAutoReset(0)
        self.simgroups={}
        for file in targetids:
            addDirProgressBar.setProgress(addDirProgressBar.progress()+1)
            app.processEvents()
            if addDirProgressBar.wasCancelled():
                addDirProgressBar.cancel()
                addDirProgressBar.done(1)
                addDirProgressBar.close(1)                                    
                return
            if file in solved: continue
            solved.append(file)
            ngid=self.generateImageID()            
            self.simgroups[ngid]=[self.meta[file]["Filename"],file,[]]            
            results=self.queryData(file,50)
            if not len(results):
                print "exception: image returned no similar img"
                continue
            for res in results:
                if res[0] in solved: continue
                if (res[1]) > (simthresd):
                    self.simgroups[ngid][2].append(res[0])
                    solved.append(res[0])
            if len(self.simgroups[ngid][2]) <= 1: del self.simgroups[ngid]
        #TODO: whats the right way of making sure this damn progressdialogs disappear ?
        addDirProgressBar.cancel()
        addDirProgressBar.done(1)
        addDirProgressBar.close(1)
        self.changedDB({"scope":"SimGroup","reason":"updated"})
    def hashStr(self,str):
        """returns MD5 hex representation """
        return md5.new(str).hexdigest()
    def getThumbName(self,fname):
        """return the thumbnail full path for the image given by fname (fname is full path) """
        return self.thdir2+ md5.new("file://"+fname).hexdigest()+".png"
    def getThumb(self,fname,imgDims=None,force=0): # TODO: imgDims is deprecated
        """use to get a QPixmap with the thumbnail for a file

        Using the standard from http://triq.net/~pearl/thumbnail-spec/
        
        @param fname: full path for image
        @type fname: string
        @param force: if true, thumbnail will be regenerated (ie, cached copy will be ignored and rewritten)
        @type force: int
        @return: a QPixmap() instance with the proportional thumbnail.If thummbn found on cache, open it, otherwise create one.
        """
        #TODO: not taking in consideration that fname may contain ~. I decided not to do a os.path.expanduser because that may slow down this code
        aImage=None
        thname=self.thdir2+ md5.new("file://"+fname).hexdigest()+".png"
        (base,ext) = os.path.splitext(fname)        
        fext=ext[1:].lower()
        if not os.path.exists(thname) or force:
            if fext in Settings.doc_ext:
                print "Not generating thumbnail for document file. Email \"imgseek-devel@lists.sourceforge.net\" if you think imgSeek should thumbnail documents."
                return None
            if fext in Settings.qt_ext:            
                aImage=QImage()
                if not aImage.load(fname):
                    print "Error loading image file while creating a thumbnail for it:" + fname
                    return None
                if qtv>=3:
                    aImage=aImage.smoothScale(128,128,QImage.ScaleMin)
                else:
                    nw=128
                    nh=128
                    if aImage.width()>aImage.height():
                        nw=128
                        nh=int(128*aImage.height()/aImage.width())
                    else:
                        nh=128
                        nw=int(128*aImage.width()/aImage.height())
                    aImage=aImage.smoothScale(nw,nh)
                aImage.save(thname,"PNG")
            elif Settings.hasImMagick and (fext in Settings.magick_ext):
                imgdb.magickThumb(fname,thname)
            elif Settings.hasPIL and (fext in Settings.pil_ext):
                im = Image.open(fname)
                nw=128
                nh=128
                sz=im.size
                if sz[0] > sz[1]:
                    nw=128
                    nh=int(128*sz[1]/sz[0])
                else:
                    nh=128
                    nw=int(128*sz[0]/sz[1])
                im=im.resize((nw,nh))
                im.save(thname)
                del im
        if not aImage: # load it from thumbnail file
            aImage=QPixmap()
            if not aImage.load(thname):
                print "Error loading thumbnail file " + thname
        else:
            aPix=QPixmap()
            aPix.convertFromImage(aImage)
            return aPix
        return aImage
    def addFile(self,fname,newid,dirid,mindim=0):
        """adds this filename to database.

        First, loading it, then calculating its haar transform
        and then adding the image index to all respective buckets.
        Each self.files list element is [filename,avg lum]

        @param fname: full filename
        @type fname: string
        @param newid: id this file should be assigned to
        @param dirid: id of the dir this file belongs to
        @param mindim: if this file has dimensions smaller than mindim, it's not added. (width and height will be checked separately)
        @type mindim: int
        
        """
        if self.dirs.has_key(dirid):
            if fname in self.dirs[dirid][4]:
                return 2
        thname=self.getThumbName(fname)
        if qtv==2:
            ext=self.extIsImg(fname)
            if ext in Settings.qt_ext:
                ret=imgdb.addImage(newid,fname,thname,not os.path.exists(thname),mindim)                
            else:
                im = Image.open(fname)
                im.save(self.thdir2+".cachev.bmp")
                del im
                ret=imgdb.addImage(newid,self.thdir2+".cachev.bmp",thname,not os.path.exists(thname),mindim)                
        else:
            ret=imgdb.addImage(newid,fname,thname,not os.path.exists(thname),mindim)            
        if not ret:
            print "Error adding image:",fname
            return 0
        if ret==2:                      # dimension too small or too white
            print "Ignored (small dimensions): "+fname
            return 0
        imgDims=[imgdb.getImageWidth(newid),imgdb.getImageHeight(newid)]
        ## init metadata
        try:
            self.meta[newid]=self.blankMetaDict.copy()
            try:
                self.initImgMetadata(fname,newid,imgDims=imgDims)
            except:
                traceback.print_exc()
                print "Error setting basic metadata for "+fname
                self.meta[newid]["Filename"]=fname
        except:
            print "Error gathering metadata for ",fname
            traceback.print_exc()
        return 1
    def initImgMetadata(self,fname,newid,imgDims=None):
        """add basic metadata to img

        @param newid: image id
        @param fname: image fullpath fname, used to extract short filename
        @param imgDims: image dimension
        @type imgDims: tuple (width,height)

        """
        self.meta[newid]=self.blankMetaDict.copy()
        try:
            self.meta[newid]["Filename"]=os.path.split(fname)[-1]
            self.meta[newid]["Filesize"]=self.prettysize(os.stat(fname).st_size)
            self.meta[newid]["Format"]=self.getImageFormat(fname)
            self.meta[newid]["Database date"]=time.asctime(time.localtime())
            self.meta[newid]["Modify date"]=time.ctime(os.stat(fname).st_mtime)
            if imgDims:
                self.meta[newid]["Dimensions"]=str(imgDims[0])+" x "+str(imgDims[1])                
        except:
            traceback.print_exc()
            print "Error setting basic metadata for "+fname
            self.meta[newid]["Filename"]=fname
        
    def prettysize(self,uint):
        """returns a string after converting the integer uint and adding periods to enhance readability

        @param uint: int to be transformed
        @type uint: int
        @return: pretty string (human readable)
        @rtype: string
        """
        a=str(uint)
        b=""
        for i in range(len(a)):
            if not i%3:
                b="."+b
            b=a[len(a)-1-i]+b
        return b[:-1] + " bytes"
    def addStubDir(self,path,volid):
        """ make sure this path exists on db

        if it already exists, return its id, otherwise, create new dir and return new id

        @param path: full dir path
        @param volid: volume id this dir belongs to
        """
        ndid=-1
        for dirid in self.volumes[volid][0]:        
            if self.dirs[dirid][0]==path:
                ndid=dirid                 # this dir is already on dbase
                break
        if ndid == -1:                  # its a new dir
            ndid=self.generateImageID()
            # guess parent
            parent=-1
            for dirid2 in self.volumes[volid][0]:
                if path[:rfind(path[:-1],os.sep)+1]==self.dirs[dirid2][0]:
                    parent=dirid2
                    break
            self.dirs[ndid]=[path,parent,[],[],[],"",volid]
            if parent != -1:
                self.dirs[parent][2].append(ndid)
            self.volumes[volid][0].append(ndid)            
        return ndid
    def addDir(self,volid,path,groupid,aborted=[],recursive=0,calledRecurs=-1,igntext="",minsize=None,pb=None,mindim=0,ignext=Settings.doc_ext,exexif=0,exiptc=0,probeext=0):
        """Add to db all the files on this dir.
        
        aborted is an empty list that should be nonempty when you want this adddir call to be aborted
        calledRecurs is the id of the dir that should be the parent to the one at path
        igntext and minsize and mindim are file criteria to ignore
        pb is a QProgressDialog
        TODO: all addDir options should be a dictionary and a dict instance should be passed as the only parm
        """
        if aborted: return -1
        #startt=time.time()
        dfiles=os.listdir(path)
        if not len(dfiles):
            return -1
        if path[-1] != os.sep[0]: path=path+os.sep
        myLabel=path
        mySteps=len(dfiles)
        hasImage=0
        if self.wnd:                    # init progress bars
            if calledRecurs==-1:
                addDirPB=self.wnd.progressBar
                self.wnd.progressList.clear()
                self.wnd.progressList.insertItem("Started adding, please wait...")
            else:
                self.wnd.progressLabel.setText(myLabel)
                addDirPB=self.wnd.subprogressBar
            addDirPB.setTotalSteps(mySteps)
            myStep=0
            addDirPB.setProgress(myStep)        
        ndid=-1
        childdirs=[]
        childdirstofix=[]
        parts=split(path,os.sep)[1:-1]
        cparts=[]
        acc=""
        for part in parts:
            acc=acc+part+os.sep
            cparts.append(os.sep+acc[:])
        for part in cparts[:-1]:
            self.addStubDir(part,volid)
        ndid=self.addStubDir(cparts[-1],volid)
        if ndid == -1:                  # its a new dir
            print "[ERROR] New dir should have an id"
            ndid=self.generateImageID()
        childfiles=[]
        childfilenames=[]
        for file in dfiles:             # note: file is not full path. Use path+file when needed
            if self.wnd:
                addDirPB.setProgress(addDirPB.progress()+1)
                if self.app: self.app.processEvents()
                if self.wnd.abortedAdd:
                    aborted.append(1)
                    self.wnd.abortedAdd=0
                    break            
            if os.path.isdir(path+file) and recursive:
                if file[0]=='.': continue # do not add dirs starting with . (also excludes .thumbnail dirs)
                # update paths tree                
                myStep=addDirPB.progress()
                thasimg=self.addDir(volid,path+file,groupid,aborted,recursive,ndid,igntext,minsize,ignext=ignext,exiptc=exiptc,exexif=exexif,probeext=probeext)
                if self.wnd:
                    self.wnd.progressLabel.setText(myLabel)
                    addDirPB.setTotalSteps(mySteps)
                    addDirPB.setProgress(myStep)
                if thasimg == -1: continue
                childdirs.append(thasimg)
                hasImage=1
                continue
            # is file
            ext=self.extIsImg(path+file,probe=probeext)
            if ext: #its a file, now just check if its an image
                if file[0]=='.': continue # do not add imgs starting with . (also excludes .thumbnail dirs)
                if igntext and find(path+file,igntext) != -1: continue
                if ignext and ext in ignext: continue
                try:
                    if minsize and os.stat(path+file).st_size < minsize*1024: continue
                except:
                    # what the hell am I supposed to do with a file that cant even be stat'd ?
                    continue
                nfid=self.generateImageID()
                self.img[nfid]=[path+file,ndid,[groupid]] # add img to dbase
                try:
                    ret=self.addFile(path+file,nfid,ndid,mindim)
                except:
                    traceback.print_exc()
                    print "Unhandled error while adding file. Please report this bug to \"imgseek-devel@lists.sourceforge.net\""
                if ret==1:
                    hasImage=1
                    childfiles.append(nfid)
                    childfilenames.append(path+file)
                    self.groups[groupid][3].append(nfid)
                    ### extract optional metadata
                    if (ext in Settings.exif_ext) and exexif:
                        try:
                            self.extractEXIF(nfid)
                        except:
                            traceback.print_exc()                            
                            print "Error extracting EXIF metadata for:",path+file
                    if (ext in Settings.iptc_ext) and Settings.hasIPTC and exiptc:
                        try:
                            self.extractIPTC(nfid)
                        except:
                            traceback.print_exc()
                            print "Error extracting IPTC metadata for:",path+file
                elif not ret: #error adding image or image already on dbase
                    if self.wnd: self.wnd.progressList.insertItem("Unable to add: "+path+file)
                    del self.img[nfid]
                elif ret==2:
                    del self.img[nfid]
        if hasImage: # only add to databse if has image isinde            
            if ndid not in self.volumes[volid][0]:self.volumes[volid][0].append(ndid)
            if ndid in self.dirs.keys():
                for ddd in childdirs:
                    while ddd in self.dirs[ndid][2]:
                        self.dirs[ndid][2].remove(ddd)                        
                self.dirs[ndid][2]=self.dirs[ndid][2]+childdirs
                self.dirs[ndid][3]=self.dirs[ndid][3]+childfiles
                self.dirs[ndid][4]=self.dirs[ndid][4]+childfilenames
                self.dirs[ndid][6]=volid
            else:
                print "Dir should already be in dbase"
                self.dirs[ndid]=[path,calledRecurs,childdirs,childfiles,childfilenames,"",volid]
            for dirid in childdirs:     # make sure all child dirs has the new dir as their parent. This is needed on the correction made
                                        # on the beginning, where a parent is added to dbase after the child. If this isn't done, the child
                                        # that was already on dbase, will still has another dir (almos always -1 (root on volume) as a parent
                try:
                    self.dirs[dirid][1]=ndid
                except:
                    pass
        self.dirty=1
        #print "Adding took %f" % (time.time()-startt)
        if self.wnd:        
            if calledRecurs==-1:
                self.wnd.progressList.insertItem("Finished successfully.")
        if hasImage: return ndid
        else: return -1
    def extractEXIF(self,fid):
        """hook for exif library.

        @param fid: set metadata for the image with this id.
        """
        try:
            file=open(self.img[fid][0], 'rb')
        except:
            print 'Unable to open file in order to extract EXIF data.'
            return
        data=EXIF.process_file(file)
        if data:
            x=data.keys()
            for i in x:
                if i in ('JPEGThumbnail', 'TIFFThumbnail'):
                    continue
                try:
                    self.meta[fid][i]=data[i].printable
                except:
                    try:
                        self.meta[fid][i]=data[i]
                    except:
                        print "Error extracting EXIF field:",i
            if data.has_key('JPEGThumbnail'):
                self.meta[fid]['JPEGThumbnail']="True"
            else:
                self.meta[fid]['JPEGThumbnail']="False"            
    def extractIPTC(self,fid):
        """hook for iptc PIL library.

        @param fid: set metadata for the image with this id.
        """
        try:
            data=IptcExtract.getiptcinfo(self.img[fid][0])
        except:
            traceback.print_exc()
            print 'Error extracting IPTC data.'
            return
        if not data:
            return
        else:
            for k, v in data.items():
                try:
                    if not k[1]: continue # skip key 0, which is \x00\x02 for example. I don't know what it means
                    i=IptcExtract.infomap[k[1]]
                    self.meta[fid][i]=str(v)
                except:
                    print "----------\nError extracting IPTC field: %s %s\nIf you know what this field means according to IPTC standards, please tell us at \"imgseek-devel@lists.sourceforge.net\"\n----------"% (k, repr(v))
    def getTotalFiles(self):
        """ number of images in db

        @return: number of images in db
        @rtype: int
        """
        return len(self.img.keys())
    def TextReport(self):
        """get database statistics

        @return: human readable statistics about db
        @rtype: string
        """
        ntext="Database status:\n"
        ntext=ntext+"Images: %d\n"%len(self.img.keys())
        ntext=ntext+"Volumes: %d\n"%len(self.volumes.keys())
        ntext=ntext+"Directories: %d\n"%len(self.dirs.keys())
        ntext=ntext+"Groups: %d\n"%len(self.groups.keys())
        ntext=ntext+"Work Batches: %d\n"%len(self.batches.keys())
        ntext=ntext+"Similarity groups: %d\n"%len(self.simgroups.keys())
        ntext=ntext+"Metadata fields cached: %d\n"%len(self.metafields)
        ntext=ntext+"Text query history entries: %d\n"%len(self.textqueryhistory)
        ntext=ntext+"Database size: %d bytes\n"%self.kbsize
        return ntext
    def newGroup(self,text,pargid=-1):
        """create new group

        @param text: new name
        @param pargid: parent group id
        @return: new group id
        """
        nid=self.generateImageID()
        self.groups[nid]=[text,"",pargid,[],[],nid]
        if pargid != -1:
            # tell parent he has a new child
            self.groups[pargid][4].append(nid)
        self.dirty=1
        self.changedDB({"scope":"Group","reason":"new","subject":nid})        
        return nid
    def removeGroup(self,gid):
        """remove group from database

        child imgs will be moved to Orphan group
        
        @param gid: group id to be removed        
        """
        self.doNotSave=1
        if gid==1:
            print "Attempt to remove default Orphan group, aborting remove process."
            return
        # move child imgs to default orphan group (id=1)
        try:
            for id in self.groups[gid][3]:
                self.moveFileGroup(id,1)
        except:
            print "Strange error removing group and adding children to Orphan group"
            traceback.print_exc()
        # tell parent i died
        if self.groups[gid][2] != -1:
            try:
                self.groups[self.groups[gid][2]][4].remove(gid)
            except:
                traceback.print_exc()
                print "Error removing group from it's parent."                
        # remove children groups
        for sid in self.groups[gid][4]:
            self.removeGroup(sid)
        del self.groups[gid]
        self.doNotSave=0
        self.changedDB({"scope":"Group","reason":"removed","subject":gid})
        self.dirty=1
    def removeVolume(self,gid):
        """remove volume and all it's dirs

        @param gid: volume id
        """
        self.doNotSave=1        
        for sid in self.volumes[gid][0]:
            self.removeDir(sid)
        del self.volumes[gid]
        self.dirty=1
        self.doNotSave=0        
        self.changedDB({"scope":"Volume","reason":"removed","subject":gid})
    def removeDir(self,gid):
        """remove dir and all it's images

        @param gid: dir id
        """        
        #remove child dirs
        self.doNotSave=1        
        for sid in self.dirs[gid][2]:
            self.removeDir(sid)
        #remove child imgs
        try:
            for sid in self.dirs[gid][3]:            
                self.removeFile(sid)                
        except:
            traceback.print_exc()        
        #tell parent dir im no longer a child
        try:
            if self.dirs[gid][1] != -1:
                self.dirs[self.dirs[gid][1]][2].remove(gid)
        except:
            traceback.print_exc()
        #remove from volume
        try:
            self.volumes[self.dirs[gid][6]][0].remove(gid)
        except:
            traceback.print_exc()        
        #remove me !!!        
        del self.dirs[gid]
        self.doNotSave=0        
        self.changedDB({"scope":"Dir","reason":"removed","subject":gid})
        self.dirty=1
    def newVolume(self,txt):
        """ create a new volume with the provided name and return its id. -1 is returned on a failure"""
        for grp in self.volumes.keys():
            if self.volumes[grp][3]==txt:
                return -1
        nid=self.generateImageID()
        while nid in self.volumes.keys():
            nid=self.generateImageID()
        self.volumes[nid]=[[],"","No description given.",txt]
        self.changedDB({"scope":"Volume","reason":"new","subject":nid})                
        return nid
    def renameVolume(self,vid,text):
        """ rename volume

        @param vid: volume id to be changed
        @param text: new volume name
        """
        self.volumes[vid][3]=text
        self.dirty=1
        self.changedDB({"scope":"Volume","reason":"renamed","subject":vid})
    def renameGroup(self,vid,text):
        self.groups[vid][0]=text
        self.dirty=1
        self.changedDB({"scope":"Group","reason":"renamed","subject":vid})        
    def describeVolume(self,vid,text):
        self.volumes[vid][2]=text
        self.dirty=1
        self.changedDB({"scope":"Volume","reason":"described","subject":vid})        
    def describeGroup(self,vid,text):
        self.groups[vid][1]=text
        self.dirty=1
        self.changedDB({"scope":"Group","reason":"described","subject":vid})        
    def removeFile(self,fid):
        imgdb.removeID(fid)
        for did in self.dirs.keys():
            try:
                self.dirs[did][3].remove(fid)
                self.dirs[did][4].remove(self.img[fid][0])
            except:
                pass
        for did in self.groups.keys():
            try:
                self.groups[did][3].remove(fid)
            except:
                pass
        for did in self.simgroups.keys():
            try:
                self.simgroups[did][2].remove(fid)
                if self.simgroups[did][1]==fid:
                    del self.simgroups[did]
            except:
                pass
        try:
            del self.meta[fid]
        except:
            pass
        try:
            del self.img[fid]
        except:
            pass
        self.changedDB({"scope":"Img","reason":"removed","subject":fid})        
        self.dirty=1
    def moveGroup(self,fromid,toid):
        """ call to move groups

        @param fromid: id of the group that will be moved
        @param toid: id of the group that will become the parent of fromid
        """
        self.doNotSave=1        
        if fromid==1:
            print "Attempt to move Orphan group. Ignored request"
            return
        if fromid==toid:
            return 0
        if toid in self.groups[fromid][4]:
            print "Invalid move. Parent -> Child"
            return 0
        if self.isGroupAncestor(toid,fromid):
            print "Attempt to move/copy parent to a child, ignoring request."
            return 0
        if self.groups[fromid][2] != -1:
            self.groups[self.groups[fromid][2]][4].remove(fromid)
        self.groups[fromid][2]=toid
        if toid != -1: self.groups[toid][4].append(fromid)
        self.doNotSave=0        
        self.changedDB({"scope":"Group","reason":"moved","subject":fromid,"target":toid})        
        self.dirty=1
        return 1
    def addImageToGroup(self,id,gid):
        """append image to a group

        @param id: image id
        @param gid: group id. This image will now also belong to this group
        """
        if id not in self.groups[gid][3]:
            self.groups[gid][3].append(id)
        if 1 not in self.img[id][2]:         # update img's parent group id list
            self.img[id][2].append(1)
        self.changedDB({"scope":"Group","reason":"addedimage","subject":gid,"target":id})            
    def moveFileGroup(self,fromid,toid,fromgroup=-1):
        """ call to make file fromid to be a child of group toid
        if fromgroup is defined, erase this image from the group fromgroup
        """
        if fromid==toid:
            return
        if toid==-1:
            print "Invalid move: Only groups may be copied/moved to root. Ignoring request."
            return
        if 0:
            for fromgid in self.img[fromid][2]: # for every group this img is in
                try:
                    self.groups[fromgid][3].remove(fromid)
                except:
                    pass               
                try:
                    self.img[fromid][2].remove(fromgid)
                except:
                    traceback.print_exc()
                    pass        
        if fromgroup!=-1:
            try:
                self.groups[fromgroup][3].remove(fromid)
            except:
                pass
            try:
                self.img[fromid][2].remove(fromgroup)
            except:
                pass
        if fromid not in self.groups[toid][3]:
            self.groups[toid][3].append(fromid)            
        self.dirty=1
        self.changedDB({"scope":"Img","reason":"movedgroup","subject":fromid,"target":toid,"fromgroup":fromgroup})        
    def crawlVolumeForDir(self,vid,lst):
        """crawl this volume adding to lst all dir id's found"""
        for dirid in self.volumes[vid][0]:
            self.crawlDirForDir(dirid,lst)
    def crawlDirForDir(self,dirid,lst):
        """crawl this volume adding to lst all dir id's found"""
        for chid in self.dirs[dirid][2]:
            self.crawlDirForDir(chid,lst)
        lst.append(dirid)
    def crawlVolumeForImg(self,vid,lst):
        """crawl this volume adding to lst all img id's found"""
        for dirid in self.volumes[vid][0]:
            self.crawlDirForImg(dirid,lst,0)
    def crawlDirForImg(self,dirid,lst,recurse=1):
        """crawl this volume adding to lst all img id's found"""
        if recurse:
            for chid in self.dirs[dirid][2]:
                self.crawlDirForImg(chid,lst)
        for chid in self.dirs[dirid][3]:
            lst.append(chid)
    def crawlGroupForImg(self,dirid,lst):
        """crawl this volume adding to lst all img id's found"""
        for chid in self.groups[dirid][4]:
            self.crawlGroupForImg(chid,lst)
        for chid in self.groups[dirid][3]:
            lst.append(chid)
    def crawlVolumeForFile(self,vid,lst):
        """crawl this volume adding to lst all img id's found"""
        for dirid in self.volumes[vid][0]:
            self.crawlDirForFile(dirid,lst,0)
    def crawlDirForFile(self,dirid,lst,recurse=1):
        """crawl this volume adding to lst all img id's found"""
        if recurse:
            for chid in self.dirs[dirid][2]:
                self.crawlDirForFile(chid,lst)
        for chid in self.dirs[dirid][3]:
            lst.append(self.img[chid][0])
    def crawlSysDirForFile(self,dirid,lst,recurse=1):
        """crawl this volume adding to lst all img id's found"""
        if dirid[-1]!='/':dirid=dirid+'/'
        dfiles=os.listdir(dirid)
        for file in dfiles:
            if os.path.isdir(dirid+file) and recurse:
                self.crawlSysDirForFile(dirid+file,lst,recurse)
                continue            
            ext=self.extIsImg(file)
            if ext: #its a file, now just check if its an image
                if file[0]=='.': continue # do not add imgs starting with . (also excludes .thumbnail dirs)
                lst.append(dirid+file)
    def crawlGroupForFile(self,dirid,lst):
        """crawl this volume adding to lst all img id's found"""
        for chid in self.groups[dirid][4]:
            self.crawlGroupForFile(chid,lst)
        for chid in self.groups[dirid][3]:
            lst.append(self.img[chid][0])            
    def crawlBatchForImg(self,bid,idlist,dumpids=1):
        """ here bid is the a batch id of a batch in db
        if dumpids==1 it will dump ids, otherwise it will dump filenames (fullpaths)
        """
        for it in self.batches[bid][1]:
            if it[0]=="Volume":
                self.crawlVolumeForImg(it[1],idlist)
            if it[0]=="Dir":
                self.crawlDirForImg(it[1],idlist)
            if it[0]=="Group":
                self.crawlGroupForImg(it[1],idlist)
            if it[0]=="Img":
                if dumpids:
                    idlist.append(it[1])
                else:
                    idlist.append(self.img[it[1]][0])
    def crawlBatchForFile(self,bid,idlist):
        """ here bid is the a batch id of a batch in db
        will dump on idlist all image files
        """
        for it in self.batches[bid][1]:
            if it[0]=="Volume":
                self.crawlVolumeForFile(it[1],idlist)
            if it[0]=="Dir":
                self.crawlDirForFile(it[1],idlist)
            if it[0]=="SysDir":
                self.crawlSysDirForFile(it[1],idlist)
            if it[0]=="Group":
                self.crawlGroupForFile(it[1],idlist)
            if it[0]=="Img":
                idlist.append(self.img[it[1]][0])
            if it[0]=="File":
                idlist.append(it[1])
            
    def crawlBatchContentsForImg(self,bid,idlist):
        """here bid is a list of items, with the same format a batch content would have """
        for it in bid:
            if it[0]=="Volume":
                self.crawlVolumeForImg(it[1],idlist)
            if it[0]=="Dir":
                self.crawlDirForImg(it[1],idlist)
            if it[0]=="Group":
                self.crawlGroupForImg(it[1],idlist)
            if it[0]=="Img":
                idlist.append(it[1])        
    def saveBatchMeta(self,dct,bid):
        """apply data on dict dct to the batch """
        idlist=[]
        self.crawlBatchForImg(bid,idlist)
        for id in idlist:
            for fi in dct.keys():
                self.meta[id][fi]=dct[fi]
    def BatchToText(self,bid):
        """ returns a pair of strings, which represents this batch in text form"""
        try:
            cnt=self.batches[bid][1]
        except:
            traceback.print_exc()
            print "Attempt to show invalid batch"
            return
        ret=[]
        for it in cnt:
            tp="Unknown"
            tx=""
            if it[0] == "SysDir":
                tp="System directory"
                tx=it[1]
            if it[0] == "Dir":
                tp="Database directory"
                tx= "%s - %d image(s)" % (self.dirs[it[1]][0],len(self.dirs[it[1]][3]))
            if it[0] == "Group":
                tp="Group"
                tx= "%s - %d image(s)" % (self.groups[it[1]][0],len(self.groups[it[1]][3]))
            if it[0] == "SimGroup":
                tp="Similarity Group"
                tx= "%s - %d image(s)" % (self.img[self.simgroups[it[1]][1]][0],len(self.simgroups[it[1]][2]))
            if it[0] == "Img":
                tp="Database image"
                tx= "%s - %s" % (self.img[it[1]][0],self.meta[it[1]]["Filesize"])
            if it[0] == "Volume":
                tp="Volume"
                tx= "%s - %d dirs(s)" % (self.volumes[it[1]][3],len(self.volumes[it[1]][0]))                
            if it[0] == "File":
                tp="System file"
                tx=it[1]
            ret.append([tp,tx])
        return ret
    def removeItemBatch(self,idx,bid):
        """call it to remove item idx from batch bid """
        try:
            del self.batches[bid][1][idx]
        except:
            print "Error removing item from batch"
            return
        self.changedDB({"scope":"Batch","reason":"removeditem","subject":bid,"target":idx})        
        self.dirty=1
    def addItemBatch(self,it,bid=-1):
        """call it to add item it to batch bid """
        it=[it["type"],it["id"]]
        if bid==-1: bid=self.curBatch
        try:
            if self.idInBatch(it,bid):
                return 0
            self.batches[bid][1].append(it)                        
        except:
            traceback.print_exc()
            print "Error adding  item to batch"
            return -1        
        self.changedDB({"scope":"Batch","reason":"addeditem","subject":bid,"target":it})
        return 1
    def renameFiles(self,renpairs):
        """call when files (fullpath) were physically renamed
        renpairs is dict. key is old fname value is new fname
        """
        for old in renpairs:
            self.renameFile(old,renpairs[old],warn=0)
        self.changedDB({"scope":"Batch","reason":"changeditem"})
        self.changedDB({"scope":"Img","reason":"multiplerenamed"})
        #TODO: the rename callback calls are not correct.            
    def renameFile(self,old,new,warn=1):
        """call when file (fullpath) old is renamed to new
        if warn=1,issue the proper changedDB calls
        """
        if self.fullfilenamedict.has_key(old):
            for did in self.dirs.keys():
                try:
                    self.dirs[did][4].remove(old)
                    self.dirs[did][4].append(new)
                except:
                    pass            
            self.fullfilenamedict[new]=self.fullfilenamedict[old]            
            del self.fullfilenamedict[old]
            id=self.fullfilenamedict[new]           
            self.meta[id]["Filename"]=os.path.split(new)[-1]
            self.filenamedict[self.meta[id]["Filename"]]=id
            del self.filenamedict[os.path.split(old)[-1]]            
            self.img[id][0]=new            
            if warn:
                self.changedDB({"scope":"Batch","reason":"changeditem"})
                self.changedDB({"scope":"Img","reason":"renamed","subject":id})
    def removeBatch(self,bid):
        """call it to remove batch bid """
        try:
            del self.batches[bid]
        except:
            print "Error removing batch"
            return
        self.changedDB({"scope":"Batch","reason":"removed","subject":bid})
        self.dirty=1        
    def resetBatch(self,bid):
        """call it to remove batch bid """
        try:            
            self.batches[bid][1]=[]            
        except:
            print "Error reseting batch"
            return
        self.changedDB({"scope":"Batch","reason":"reseted","subject":bid})
    def resetBatchHistory(self):
        self.batches={1:[time.asctime(time.localtime()),[]]}
        self.changedDB({"scope":"Batch","reason":"resetedbatchhistory"})        
    def addBatch(self,name=""):
        """call it to add batch  """
        bid=0
        try:
            if not name: name=time.asctime(time.localtime())
            bid=self.generateImageID()
            self.batches[bid]=[name,[]]
        except:
            traceback.print_exc()
            print "Error adding  batch"
            return 0
        self.changedDB({"scope":"Batch","reason":"new","subject":bid})
        self.dirty=1
        return bid
    def idInBatch(self,it,bid=-1):
        """checks if the given id is in one of the groups/dirs already included on a batch """
        if bid==-1: bid=self.curBatch
        for cid in self.batches[bid][1]:
            #cid=self.batches[bid][k]
            try:
                if cid[1]==it[1]:
                    if cid[0]==it[0]:
                        return 1
            except:
                traceback.print_exc()
        return 0
    def refreshDB(self,cb=None):
        count=0
        added=[]
        for vid in self.volumes.keys():
            lst=[]
            self.crawlVolumeForDir(vid,lst)
            for did in lst:                
                subdirs=[]
                for subdid in self.dirs[did][2]:  subdirs.append(self.dirs[subdid][0])
                path=self.dirs[did][0]
                if not os.path.exists(path): continue
                files=os.listdir(path)
                if cb:
                    cb.parent().app.processEvents()                                        
                    cb.setLabelText("Scanning "+path)
                    cb.setProgress(0)
                    cb.setTotalSteps(len(files))
                    cb.parent().app.processEvents()
                for file in files:
                    if cb:
                        cb.parent().app.processEvents()                    
                        cb.setProgress(cb.progress()+1)
                        cb.parent().app.processEvents()                   
                        if cb.wasCancelled():
                            cb.close()
                            return count                                    
                    if file[0]=='.': continue # do not add imgs starting with .                    
                    if os.path.isdir(path+file):
                        continue
                        #TODO: finish this, or simply remove this code. That is, the possibility to recurse when searching for new images
                        if file[-1] != os.sep: file=file+os.sep
                        if path+file in subdirs: continue
                        newid=self.addDir(vid,path+file,1,[],1,did)
                        if newid != -1:
                            self.dirs[did][2].append(newid)
                            if newid not in self.volumes[vid][0]:self.volumes[vid][0].append(newid)
                        continue
                    fiext=self.extIsImg(file)
                    if fiext not in Settings.doc_ext:
                        if path+file not in self.dirs[did][4]:
                            # add it
                            nfid=self.generateImageID()
                            self.img[nfid]=[path+file,did,None]
                            if self.addFile(path+file,nfid,did) != 1:
                                del self.img[nfid]
                                continue
                            count=count+1
                            added.append(nfid)
                            self.groups[1][3].append(nfid)
                            self.dirs[did][3].append(nfid)
                            self.dirs[did][4].append(path+file)
        self.changedDB({"scope":"Img","reason":"added","subject":added})
        self.remove_dead()
        if cb:
            cb.cancel()
            cb.hide()
            cb.close()
            del cb
        return count
    def addSysDirBookmark(self,txt):
        self.sysdirbm.append(txt)
        self.dirty=1
        for cb in self.sysdirbmCb: cb()
    def isGroupAncestor(self,chid,pid):
        """ check if pid is an ancestor of chid - to avoid infinite loops when moving/copying groups"""
        for son in self.groups[pid][4]:
            if son==chid:
                return 1
            else:
                return self.isGroupAncestor(chid,son)
    def delSysDirBookmark(self,id):
        """remove a bookmark
        
        @param id: 0-based index of the bmark to be removed
        """
        try:
            del self.sysdirbm[id]
            self.dirty=1
            for cb in self.sysdirbmCb: cb()
        except:
            traceback.print_exc()
            print "Error removing the supplied bookmark id"
    def removeAllMeta(self,id):
        for fd in self.meta[id].keys():
            if fd not in self.readonlyfields:
                del self.meta[id][fd]
        self.changedDB({"scope":"Meta","reason":"removedall","subject":id})
    def syncFilenames(self,force=0):
        """call it to update dictionary of filenames:id """
        if not force and (len(self.filenamedict.keys())==len(self.img.keys())):return # do not sync if almost nothing changed
        self.filenamedict={}
        for id in self.img.keys():
            try:
                self.filenamedict[self.meta[id]["Filename"]]=id
            except:
                print "Image without metadata found. Trying to rebuild it."
                self.initImgMetadata(self.img[id][0],id)
    def syncFullFilenames(self,force=0):
        """call it to update dictionary of filenames:id """
        if not force and (len(self.fullfilenamedict.keys())==len(self.img.keys())):return # do not sync if almost nothing changed
        self.fullfilenamedict={}
        for id in self.img.keys():
            try:
                self.fullfilenamedict[self.img[id][0]]=id
            except:
                print "Image without metadata found. Trying to rebuild it."
                self.initImgMetadata(self.img[id][0],id)                
    def queryFilename(self,fname,numres=10):
        """query for similar filenames

        @param fname: full filename
        @param numres: maximum number of results
        @return: list of similar filenames
        @rtype: list
        """
        self.syncFilenames()
        try:
            fname=os.path.split(fname)[-1]
        except:
            pass
        mats=difflib.get_close_matches(fname, self.filenamedict.keys(), numres,0.5)
        mats=mats[1:]
        fres=[]
        for mat in mats:
            fres.append([self.filenamedict[mat],0])
        return fres
    def queryImage(self,filen,numres,scanned):
        """query for similar images

        @param filen: full filename
        @param numres: maximum number of results
        @param scanned: true if image is a photo. False if it's a drawing
        @return: list of similar images. List of pairs [result id,result score]  Score is 0-100
        @rtype: list
        """
        
        imgdb.queryImgFile(filen,numres,not scanned)
        nres=imgdb.getNumResults()
        res=[]
        for i in range(nres):
            rid=imgdb.getResultID()
            rsc=imgdb.getResultScore()
            rsc=-100.0*rsc/38.70
            #sanity checks
            if rsc<0:rsc=0
            if rsc>100:rsc=100
            res.append([rid,rsc])
        res.reverse()
        if not res: return res
        return res[1:]
    def queryData(self,id,numres):
        """query for similar images. Source is a db image

        @param id: img id
        @param numres: maximum number of results
        @return: list of similar images. List of pairs [result id,result score]  Score is 0-100
        @rtype: list
        """
        
        imgdb.queryImgID(id,numres)
        nres=imgdb.getNumResults()
        res=[]
        for i in range(nres):
            rid=imgdb.getResultID()
            rsc=imgdb.getResultScore()
            rsc=-100.0*rsc/38.70
            #sanity checks
            if rsc<0:rsc=0
            if rsc>100:rsc=100
            res.append([rid,rsc])
        res.reverse()
        if not res: return res        
        return res[1:]
    def moveObjectsToGroup(self, ids,gid):
        """ call to move a group of objects to gid
        
        @param ids: a list of maps.
        @param gid: destination group
        """
        exitc=0
        self.doNotSave=1        
        try:
            for id in ids:
                if id["type"]=="Group":
                    self.moveGroup(id["id"],gid)
                if id["type"]=="Img":
                    if id.has_key("FromId"):
                        self.moveFileGroup(id["id"],gid,id["FromId"])
                    else:
                        self.moveFileGroup(id["id"],gid)
            exitc=1
        except:
            traceback.print_exc()
            exitc=1
        self.doNotSave=0
        self.changedDB({"scope":"Group","reason":"movedimages"})
        return exitc
    def copyGroup(self,sid,gid):
        if self.isGroupAncestor(gid,sid):
            print "Attempt to move/copy parent to a child, ignoring request."
            return
        nid=self.newGroup(self.groups[sid][0][:],pargid=gid)
        self.groups[nid][1]=self.groups[sid][1][:]
        self.groups[nid][3]=self.groups[sid][3][:]
        for sg in self.groups[sid][4]:
            self.copyGroup(sg,nid)
    def copyObjectsToGroup(self, ids,gid):
        """ call to copy a group of objects to gid
        ids is a list of maps. """
        exitc=0
        self.doNotSave=1
        try:
            for id in ids:
                if id["type"]=="Group":
                    self.copyGroup(id["id"],gid)
                if id["type"]=="Img":
                    self.moveFileGroup(id["id"],gid)
            exitc=1
        except:
            traceback.print_exc()
            exitc=1
        self.doNotSave=0
        self.changedDB({"scope":"Group","reason":"copiedimages"})
        return exitc
    def textquery(self,txt,numres=10):
        res=[]
        metaq=txt[0]
        logic=txt[1]
        pres=[]
        for fi in range(len(metaq)):
            pres.append([])
        for fil in self.meta.keys():            
            for fi in range(len(pres)):
                try:
                    field=metaq[fi][0]
                    if not self.meta[fil].has_key(field) or not self.meta[fil][field]:
                        continue
                    else:
                        if find(lower(self.meta[fil][field]),lower(metaq[fi][1])) != -1:
                            pres[fi].append(fil)
                            
                except:
                    traceback.print_exc()
                    pass
        if logic=="AND":
            min=0
            for fi in range(len(pres)):
                if len(pres[fi]) < len(pres[min]):
                    min=fi
            for id in pres[min]:
                failed=0
                for fi in range(len(pres)):
                    if id not in pres[fi]:                        
                        failed=1
                        break
                if not failed and id not in res: res.append(id)                
        elif logic=="OR":
            for fi in range(len(pres)):
                for id in pres[fi]:
                    if id not in res: res.append(id)
        fres=[]
        for id in res:
            fres.append([id,0])
            if len(fres)>=numres:
                return fres
        return fres
    def closedb(self):
        self.savedb()
        self.dirty=0
        imgdb.closeDbase()
    def regenerateThumbs(self):
        for id in self.img.keys():
            self.getThumb(self.img[id][0],force=1)
    def importMeta(self,header,data,key,fullpath=1):
        """ header is a list of meta fields (strings
        data is a list of lists of strings
        if fullpath==1: match key field to full path instead of filename
        key is index of column to match when looking which file in db to set this data
        """
        resp=""
        count=0
        self.syncFilenames()        
        for it in data:
            try:
                id=-1
                if fullpath:
                    for itd in self.img.keys():
                        if self.img[itd][0]==it[key]:
                            id=itd
                            break
                else:
                    if self.filenamedict.has_key(it[key]):
                        id=self.filenamedict[it[key]]
                if id==-1:
                    continue
                for hid in range(len(header)):
                    if header[hid] in self.readonlyfields: continue
                    if hid==key: continue
                    self.meta[id][header[hid]]=it[hid]
                count=count+1
                self.changedDB({"scope":"Meta","reason":"changed","subject":id})        
            except:
                traceback.print_exc()
                continue
        resp=resp+"%d item(s) imported."%count
        return resp
    def openImage(self,fname):
        """abstract method for opening an image. This function should call whatever is available to open the given image

        @param fname: full path for an image
        @return QImage instance
        """
        (base,ext) = os.path.splitext(fname)        
        fext=ext[1:].lower()        
        aImage=QImage()
        if fext in Settings.qt_ext:
            if not aImage.load(fname):
                print "Error loading image file " + fname
        elif Settings.hasImMagick and (fext in Settings.magick_ext):
            imgdb.convert(fname,self.thdir2+".cachev.bmp")
            if not aImage.load(self.thdir2+".cachev.bmp"):
                print "Error loading cached image file " + fname
        elif Settings.hasPIL and (fext in Settings.pil_ext):
            im = Image.open(fname)
            im.save(self.thdir2+".cachev.bmp")
            del im
            if not aImage.load(self.thdir2+".cachev.bmp"):
                print "Error loading cached image file " + fname                
        return aImage
    def openPixmap(self,fname):
        """abstract method for opening an image. This function should call whatever is available to open the given image

        @param fname: full path for an image
        @return QPixmap instance
        """
        (base,ext) = os.path.splitext(fname)        
        fext=ext[1:].lower()        
        aImage=QPixmap()
        if fext in Settings.qt_ext:
            if not aImage.load(fname):
                print "Error loading image file " + fname
        elif Settings.hasImMagick and (fext in Settings.magick_ext):
            imgdb.convert(fname,self.thdir2+".cachev.bmp")
            if not aImage.load(self.thdir2+".cachev.bmp"):
                print "Error loading cached image file " + fname
        elif Settings.hasPIL and (fext in Settings.pil_ext):
            im = Image.open(fname)
            im.save(self.thdir2+".cachev.bmp")
            del im
            if not aImage.load(self.thdir2+".cachev.bmp"):
                print "Error loading cached image file " + fname                
        return aImage
        
    def scanForMetaTags(self,bid):
        """rescan all images in this batch for metadata on the files """
        ids=[]
        self.crawlBatchForImg(bid,ids)
        for nfid in ids:
            fname=self.img[nfid][0]            
            try:
                if not self.meta.has_key(nfid): self.meta[nfid]=self.blankMetaDict.copy()
                try:
                    self.initImgMetadata(fname,nfid)
                except:
                    traceback.print_exc()
                    print "Error setting basic metadata for "+fname
                    self.meta[nfid]["Filename"]=fname
            except:
                print "Error gathering metadata for ",fname
                traceback.print_exc()
            ### extract optional metadata
            ext=self.extIsImg(fname,probe=1)
            if (ext in Settings.exif_ext):
                try:
                    self.extractEXIF(nfid)
                except:
                    traceback.print_exc()                            
                    print "Error extracting EXIF metadata for:",fname
            if (ext in Settings.iptc_ext) and Settings.hasIPTC:
                try:
                    self.extractIPTC(nfid)
                except:
                    traceback.print_exc()
                    print "Error extracting IPTC metadata for:",fname
        
    def createGroupsFromSymGroups(self):
        self.doNotSave=1
        for sid in self.simgroups.keys():
            ngid=self.newGroup("["+self.meta[self.simgroups[sid][1]]["Filename"]+"]")
            self.groups[ngid][1]="Images similar to "+self.img[self.simgroups[sid][1]][0]
            for imid in self.simgroups[sid][2]:
                self.groups[ngid][3].append(imid)
        self.doNotSave=0
        self.changedDB({"scope":"Group","reason":"added"})
        
